package dfki.protege.flora_tab;

import java.awt.*;
import java.awt.dnd.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import edu.stanford.smi.protege.widget.*;
import edu.stanford.smi.protege.resource.*;
import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.ui.*;
import edu.stanford.smi.protege.util.*;


public class FloraTab extends AbstractWidget {

    private InstanceDisplay itsInstanceDisplay;
    private InstanceClsesPanel itsClsesPanel;
    private DirectInstancesList itsDirectInstancesList;
    private JTextArea dribbleTextArea;
    private InstancesList instancesList;
    private Flora flora;
    private String kbDirectory;
    private String xsbCommand;
    private boolean stderr;
    private String kbFileName;

    public void initialize() {
        // setIcon(Icons.getInstancesIcon()); // query icon!!!
        setLabel("Flora");
        add(createQuerySplitter());
        transmitSelection();
        setupDragAndDrop();
	kbDirectory = ApplicationProperties.getString("flora.directory", "");
	xsbCommand = 
	  ApplicationProperties.getString("flora.xsb.command", "xsb");
	stderr = 
	  ApplicationProperties.getString("flora.xsb.stderr", "true")
	    .startsWith("t");
        flora = new Flora(xsbCommand, stderr, dribbleTextArea);
        kbFileName = ExportFlora.normalize(kbDirectory + 
          getKnowledgeBase().getName() + ".flr");
    }


    // ======================================================================
    // parts added for queries [MS]
    // ======================================================================

    private JComponent createQuerySplitter() {
        JSplitPane pane = createTopBottomSplitPane("FloraTab.top_down", 250);
        pane.setTopComponent(createClsSplitter());
        pane.setBottomComponent(createQueryPanel());
        return pane;
    }


    private JComponent createQueryPanel() {

        JSplitPane resultDribblePane = 
          createLeftRightSplitPane("FloraTab.bottom.left_right", 250);

        dribbleTextArea = new JTextArea();
	dribbleTextArea.setEditable(false);
        LabeledComponent dribbleTextAreaComponent = 
          new LabeledComponent("Flora", new JScrollPane(dribbleTextArea));
        dribbleTextAreaComponent
          .addHeaderButton(new AbstractAction("Query", Icons.getFindIcon()) {
            public void actionPerformed(ActionEvent event) {
              Collection selection = itsDirectInstancesList.getSelection();
              if (selection.size() == 1) {
                String query = 
                  getQueryFromInstance((Instance)selection.iterator().next());
                if (query != null && !query.equals("")) {
                  try {
                    if (!flora.running()) {
                      ExportFlora.export(getProject(), kbFileName);
                      flora.start();
                      flora.loadKB(kbFileName);
                    }
                    String result = flora.query(query);
                    Collection values = ExportFlora.getResultValues(result);
                    showResultValues(values);
                  } catch (Exception e) {
                    e.printStackTrace();
                  }
                }
              } else {
                // "no query" warning ... !!!
                // or: disable query button when no query is present!!!
              }
            }
          });

        dribbleTextAreaComponent.addHeaderButton(
          new AbstractAction("Update KB", Icons.getCopyIcon()) {
            public void actionPerformed(ActionEvent event) {
              ExportFlora.export(getProject(), kbFileName);
              try {
                if (!flora.running()) {
                  flora.start();
                }
                flora.loadKB(kbFileName);
              } catch (Exception e) {
                e.printStackTrace();
              }
            }
          });
        resultDribblePane.setRightComponent(dribbleTextAreaComponent);

        instancesList = new InstancesList(getProject());
        resultDribblePane.setLeftComponent(instancesList);

        return resultDribblePane;

    }


    String getQueryFromInstance(Instance instance) {
      Collection ownSlots = instance.getOwnSlots();
      String queryTemplate = null;
      for (Iterator ownSlotIterator = ownSlots.iterator();
           ownSlotIterator.hasNext();) {
        Slot ownSlot = (Slot)ownSlotIterator.next();
        String ownSlotName = ownSlot.getName();
        if (ownSlotName.equals(ExportFlora.floraQuerySlot)) {
          // check !multiple && valueType.equals(STRING) ... !!!
          queryTemplate = instance.getOwnSlotValue(ownSlot).toString();
        }
      }
      if (queryTemplate != null && !queryTemplate.equals("")) {
        // instantiate query template
        String query = queryTemplate;
        for (Iterator templateVarIterator = ownSlots.iterator();
             templateVarIterator.hasNext();) {
          Slot slot = (Slot)templateVarIterator.next();
          String templateVar = slot.getName();
          String templateVarPattern = "<" + templateVar + ">";
          if (query.indexOf(templateVarPattern) != -1) {
            // replace by value(s)
            String valueString;
            ValueType valueType = instance.getOwnSlotValueType(slot);
            boolean multiple = instance.getOwnSlotAllowsMultipleValues(slot);
            if (multiple)
              valueString = // v1,...,vn
                ExportFlora.values(valueType, 
                  instance.getOwnSlotValues(slot), false);
            else {
              valueString = 
                ExportFlora.value(valueType, instance.getOwnSlotValue(slot));
              if (valueString == null || valueString.trim().equals(""))
                valueString = "_"; // ... ???
            }
            query = replaceAll(query, templateVarPattern, valueString);
          }
        }
        return query.toString();
      } else 
        return null;
    }

    void showResultValues(Collection values) {
      KnowledgeBase kb = getKnowledgeBase();
      Vector instances = new Vector();
      for (Iterator valueIterator = values.iterator();
           valueIterator.hasNext();) {
        String value = (String)valueIterator.next();
        Instance instance = kb.getInstance(value);
        if (instance != null)
          instances.addElement(instance);
        // else ... ???
      }
      Collections.sort(instances, new FrameComparator());
      instancesList.setInstances(instances);
    }


  static String replaceAll(String string, String from, String to) {
    // replace "from" by "to" in "string"
    int pos = string.indexOf(from);
    if (pos == -1)
      return string;
    else {
      int len = from.length();
      StringBuffer buffer = new StringBuffer(string);
      buffer.replace(pos, pos+len, to);
      return replaceAll(buffer.toString(), from, to);
    }
  }

    // ======================================================================
    // end of parts added for queries
    // ======================================================================


    private JComponent createClsSplitter() {
        JSplitPane pane = createLeftRightSplitPane("FloraTab.left_right", 250);
        pane.setLeftComponent(createClsesPanel());
        pane.setRightComponent(createInstanceSplitter());
        return pane;
    }

    private JComponent createInstanceSplitter() {
        JSplitPane pane = createLeftRightSplitPane("FloraTab.right.left_right", 250);
        pane.setLeftComponent(createInstancesPanel());
        pane.setRightComponent(createInstanceDisplay());
        return pane;
    }

    private JComponent createInstanceDisplay() {
        itsInstanceDisplay = new InstanceDisplay(getProject());
        return itsInstanceDisplay;
    }

    private JComponent createClsesPanel() {
        itsClsesPanel = new InstanceClsesPanel(getProject());
        itsClsesPanel.addSelectionListener(new SelectionListener() {
            public void selectionChanged(SelectionEvent event) {
                transmitSelection();
            }
        });
        return itsClsesPanel;
    }

    private void transmitSelection() {
        Collection selection = itsClsesPanel.getSelection();
        itsDirectInstancesList.setClses(selection);
        itsDirectInstancesList.initializeSelection();
    }

    private JComponent createInstancesPanel() {
        JPanel panel = ComponentFactory.createPanel();
        panel.setLayout(new BorderLayout());
        // panel.add(createClsDisplay(), BorderLayout.NORTH);
        panel.add(createDirectInstancesList(), BorderLayout.CENTER);
        return panel;
    }

    /*
    private JComponent createClsDisplay() {
        itsClsList = ComponentFactory.createSingleItemList(null);
        itsClsList.setCellRenderer(new FrameRenderer());
        itsClsList.addSelectionListener(new SelectionListener() {
            public void selectionChanged(SelectionEvent event) {
                Collection selection = event.getSelectable().getSelection();
                Instance firstSelection = (Instance) CollectionUtilities.getFirstItem(selection);
                itsDirectInstancesList.clearSelection();
                itsInstanceDisplay.setInstance(firstSelection);
            }
        });
        LabeledComponent c = new LabeledComponent("Class", new JScrollPane(itsClsList));
        c.setPreferredSize(new Dimension(200, 100));
        return c;
    }
    */

    private JComponent createDirectInstancesList() {
        itsDirectInstancesList = new DirectInstancesList(getProject());
        itsDirectInstancesList.addSelectionListener(new SelectionListener() {
            public void selectionChanged(SelectionEvent event) {
                Collection selection = itsDirectInstancesList.getSelection();
                Instance selectedInstance;
                if (selection.size() == 1) {
                    selectedInstance = (Instance) CollectionUtilities.getFirstItem(selection);
                } else {
                    selectedInstance = null;
                }
                // itsClsList.clearSelection();
                itsInstanceDisplay.setInstance(selectedInstance);
            }
        });
        return itsDirectInstancesList;
    }

    private void setupDragAndDrop() {
        DragSource.getDefaultDragSource().createDefaultDragGestureRecognizer(itsDirectInstancesList.getDragComponent(),
            DnDConstants.ACTION_COPY_OR_MOVE, new FloraTabDirectInstancesListDragSourceListener());
        new DropTarget(itsClsesPanel.getDropComponent(), DnDConstants.ACTION_COPY_OR_MOVE, new ClsesTreeTarget());
    }
}

class InstanceClsesPanel extends SelectableContainer {
    private Project itsProject;
    private SelectableTree itsClsTree;

    public InstanceClsesPanel(Project project) {
        itsProject = project;
        add(createClsesPanel());
        setSelectable(itsClsTree);
    }

    public JTree getDropComponent() {
        return itsClsTree;
    }

    private JComponent createClsesPanel() {
        Cls root = itsProject.getKnowledgeBase().getRootCls();
        itsClsTree = ComponentFactory.createSelectableTree(null, new ParentChildRoot(root));
        FrameRenderer renderer = new FrameRenderer();
        renderer.setDisplayDirectInstanceCount(true);
        itsClsTree.setCellRenderer(renderer);
        itsClsTree.setSelectionRow(0);
        LabeledComponent c = new LabeledComponent("Classes", new JScrollPane(itsClsTree));
        c.addHeaderButton(getViewClsAction());
        c.setFooterComponent(new ClsTreeFinder(itsProject.getKnowledgeBase(), itsClsTree, "Find Class"));
        return c;
    }

    private Action getViewClsAction() {
        return new AbstractAction("View Selected Class", Icons.getViewIcon()) {
            public void actionPerformed(ActionEvent event) {
                Iterator i = ComponentUtilities.getSelection(itsClsTree).iterator();
                while (i.hasNext()) {
                    Cls cls = (Cls) i.next();
                    itsProject.show(cls);
                }
            }
        };
    }
}

class FloraTabDirectInstancesListDragSourceListener extends ListDragSourceListener {

    public void doMove(final JComponent c, int[] indices, Collection draggedObjects) {
    }

    public void doCopy(JComponent c, int[] indices, Collection draggedObjects) {
    }
}

class ClsesTreeTarget extends TreeTarget {
    public ClsesTreeTarget() {
        super(false);
    }

    public boolean doDrop(JTree tree, Object source, int row, Object area) {
        boolean succeeded = false;

        TreePath path = tree.getPathForRow(row);
        Cls targetCls = (Cls) ((LazyTreeNode)path.getLastPathComponent()).getUserObject();
        Instance sourceInstance = (Instance) source;

        if (targetCls.isAbstract()) {
            Log.warning("cannot create instance of abstract class", this, "doDrop");
        } else if (sourceInstance.hasDirectType(targetCls)) {
            Log.trace("do nothing", this, "doDrop", source, targetCls);
        } else {
            if (sourceInstance instanceof Cls) {
                if (targetCls.isClsMetaCls()) {
                    if (confirm(tree, sourceInstance, targetCls)) {
                        sourceInstance.setDirectType(targetCls);
                        succeeded = true;
                    }
                }
            } else if (!targetCls.isClsMetaCls()) {
                if (confirm(tree, sourceInstance, targetCls)) {
                    sourceInstance.setDirectType(targetCls);
                    succeeded = true;
                }
            }
        }
        return succeeded;
    }

    private boolean confirm(final JComponent c, Instance source, Cls target) {
        boolean doChange = true;
        if (SystemUtilities.modalDialogInDropWorks()) {
            String text = "Change class of \"" + source.getBrowserText() + "\" to " + target.getName() + "?";
            int result = ModalDialog.showMessageDialog(c, text, ModalDialog.MODE_OK_CANCEL);
            doChange = result == ModalDialog.OPTION_OK;
        }
        return doChange;
    }
}


